! file: filter.f90
!
!  The file contains the filtering module with three functions
!  - kalman_filter(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0)
!  - kalman_filter_missing(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0)
!  - chand_recursion(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0)
!
!  Both of which evaluate the log likelihood of a LGSS
!
!    s_t = TT*s_{t-1} + RR*eps_t,  eps_t ~ iidN(0, QQ)
!    y_t = DD + ZZ*s_t + eta_t, eta_t ~ iidN(0, HH)
! 
!  where 
!  YY  - ny x nobs
!  TT  - ns x ns
!  RR  - ns x neps
!  QQ  - neps x neps
!  DD  - ny
!  ZZ  - ny x ns
!  HH  - ny x ny
! 
!  t0 is the number of initial observations to condition on.
!
!  The state is initialized as the stationary distribution.
!  kalman_filter_missing can handle missing data entered as `NaN.'
!
!  returns:
!  loglh -- scalar containing the log likelihood.
!
!  Helper routines:
!   - determinant(): computes determinant of matrix.
!   - dlyap(): solves discrete Lyapunov equation via SLICOT.
! 
!  In addition, the module contains the subroutine `reduce_system', which 
!   removes redundant states from the LGSS.
!
!  Gradients can be computed via:
!  - chand_recursion_derivative
!  
! Implemented by Ed Herbst <edward.p.herbst@frb.gov>.
!-------------------------------------------------------------------------------
module filter

  use mkl95_precision, only: wp => dp
  use omp_lib

  implicit none 

  real(wp), parameter :: ONE = 1.0_wp, ZERO = 0.0_wp, NEG_ONE = -1.0_wp, M_PI = 3.141592653589793_wp

  real(wp), parameter :: really_small = 1e-10
contains



  function kalman_filter(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0) result(loglh) 
    ! Evaluating the likelihood of LGSS via Kalman Filter.
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ny, nobs, neps, ns, t0
    real(wp), intent(in) :: y(:,:), TT(:,:), RR(:,:), QQ(:,:), DD(:), ZZ(:,:), HH(:,:)
    real(wp) :: loglh

    real(wp) :: At(ns), Pt(ns,ns), RQR(ns,ns), Kt(ns,ny), QQRRp(neps,ns)
    real(wp) :: yhat(ny), nut(ny), Ft(ny,ny), iFt(ny,ny), detFt
    integer :: t, info

    real(wp) :: ZZP0(ny,ns), iFtnut(ny,ny), gain(ns), C(ns,ns), KtiFt(ns,ny), TTPt(ns,ns)

    ! BLAS functions
    real(wp) :: ddot

    ! initialization 
    At = 0.0_wp
    call dgemm('n','t', neps, ns, neps, 1.0_wp, QQ, neps, RR, ns, 0.0_wp, QQRRp, neps)
    call dgemm('n','n', ns, ns, neps, 1.0_wp, RR, ns, QQRRp, neps, 0.0_wp, RQR, ns)

    call dlyap(TT, RQR, Pt, ns, info)

    ! Pt = TT*Pt*TT' + RQR
    call dgemm('n','n', ns, ns, ns, 1.0_wp, TT, ns, Pt, ns, 0.0_wp, TTPt, ns)
    Pt = RQR
    call dgemm('n','t', ns, ns, ns, 1.0_wp, TTPt, ns, TT, ns, 1.0_wp, Pt, ns)


    loglh = 0.0_wp
    do t = 1, nobs

       ! yhat = ZZ*At + DD
       call dcopy(ny, DD, 1, yhat, 1)
       call dgemv('n', ny, ns, ONE, ZZ, ny, At, 1, ONE, yhat, 1)

       ! nut = yt - yhat
       nut = y(:, t) - yhat

       ! Ft = ZZ*Pt*ZZ' + HH
       call dcopy(ny*ny, HH, 1, Ft, 1)
       call dsymm('r', 'l', ny, ns, ONE, Pt, ns, ZZ, ny, ZERO, ZZP0, ny)
       call dgemm('n', 't', ny, ny, ns, ONE, ZZP0, ny, ZZ, ny, ONE, Ft, ny)

       ! iFt = inv(Ft)
       call dcopy(ny*ny, Ft, 1, iFt, 1)
       call dpotrf('u', ny, iFt, ny, info)
       call dpotri('u', ny, iFt, ny, info)

       ! det(Ft)
       detFt = determinant(Ft, ny);

       call dsymv('u', ny, ONE, iFt, ny, nut, 1, ZERO, iFtnut, 1)

       if (t > t0) then
          loglh = loglh - 0.5_wp*ny*log(2*M_PI) - 0.5_wp*log(detFt) & 
               - 0.5_wp*ddot(ny, nut, 1, iFtnut, 1)
       endif

       ! Kt = TT*Pt*ZZ'
       call dgemm('n','t', ns, ny, ns, ONE, TT, ns, ZZP0, ny, ZERO, Kt, ns)

       ! At = TT*At + Kt*iFt*nut'
       call dgemv('n', ns, ny, ONE, Kt, ns, iFtnut, 1, ZERO, gain, 1)
       call dgemv('n', ns, ns, ONE, TT, ns, At, 1, ONE, gain, 1)
       call dcopy(ns, gain, 1, At, 1)

       ! Pt = TT*Pt*TT' + RQR - Kt*iFt*Kt'
       call dgemm('n','n', ns, ns, ns, ONE, TT, ns, Pt, ns, ZERO, C, ns)
       call dsymm('r', 'u', ns, ns, ONE, Pt, ns, TT, ns, ZERO, C, ns)
       call dcopy(ns*ns, RQR, 1, Pt, 1)
       call dgemm('n', 't', ns, ns, ns, ONE, C, ns, TT, ns, ONE, Pt, ns)

       call dsymm('r', 'u', ns, ny, ONE, iFt, ny, Kt, ns, ZERO, KtiFt, ns)
       call dgemm('n', 't', ns, ns, ny, NEG_ONE, KtiFt, ns, Kt, ns, ONE, Pt, ns)

    end do

    call mkl_free_buffers()
  end function kalman_filter




  function kalman_filter_missing(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0) result(loglh)
    ! Evaluating the log likelihood of LGSS via the Kalman Filter.
    !  Can handle missing data entered as `NaN.'
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ny, nobs, neps, ns, t0
    real(wp), intent(in) :: y(:,:), TT(:,:), RR(:,:), QQ(:,:), DD(:), ZZ(:,:), HH(:,:)
    real(wp) :: loglh

    real(wp) :: At(ns), Pt(ns,ns), RQR(ns,ns), QQRRp(neps,ns)
    real(wp) :: detFt
    integer :: t,i,j, info

    integer :: nmiss, ngood
    integer, allocatable :: oind(:)
    real(wp), allocatable :: iFtnut(:),ZZPt(:,:), Ft(:,:), iFt(:,:), Kt(:,:), KtiFt(:,:), nut(:),yhat(:)
    real(wp) :: ZZP0(ny,ns), gain(ns), C(ns,ns),  TTPt(ns,ns), HHmiss(ny,ny)

    ! BLAS functions
    real(wp) :: ddot

    ! initialization 
    At = 0.0_wp
    call dgemm('n','t', neps, ns, neps, 1.0_wp, QQ, neps, RR, ns, 0.0_wp, QQRRp, neps)
    call dgemm('n','n', ns, ns, neps, 1.0_wp, RR, ns, QQRRp, neps, 0.0_wp, RQR, ns)

    call dlyap(TT, RQR, Pt, ns, info)

    ! Pt = TT*Pt*TT' + RQR

    call dgemm('n','n', ns, ns, ns, 1.0_wp, TT, ns, Pt, ns, 0.0_wp, TTPt, ns)
    Pt = RQR
    call dgemm('n','t', ns, ns, ns, 1.0_wp, TTPt, ns, TT, ns, 1.0_wp, Pt, ns)


    loglh = 0.0_wp

    do t = 1, nobs
       HHmiss = 0.0_wp

       nmiss = count(isnan(y(:,t)))

       ngood = ny - nmiss
       allocate(oind(ngood),iFtnut(ngood),ZZPt(ngood, ns),Ft(ngood, ngood),iFt(ngood,ngood),Kt(ns,ngood), KtiFt(ns,ngood),nut(ngood),yhat(ngood))
       j = 1;
       do i = 1,ny
          if (isnan(y(i,t))) then 
          else
             oind(j) = i
             j = j + 1;
          end if
       end do


       ! yhat = ZZ*At + DD
       call dcopy(ngood, DD(oind), 1, yhat, 1)
       call dgemv('n', ngood, ns, ONE, ZZ(oind,:), ngood, At, 1, ONE, yhat, 1)

       ! nut = yt - yhat
       nut = y(oind, t) - yhat

       ! Ft = ZZ*Pt*ZZ' + HH
       call dcopy(ngood*ngood, HH(oind, oind), 1, Ft, 1)

       !call dsymm('r', 'l', ngood, ns, ONE, Pt, ns, ZZ(oind,:), ngood, ZERO, ZZPt, ngood)
       call dgemm('n','n', ngood, ns, ns, ONE, ZZ(oind, :), ngood, Pt, ns, ZERO, ZZPt, ngood)
       call dgemm('n', 't', ngood, ngood, ns, ONE, ZZPt, ngood, ZZ(oind,:), ngood, ONE, Ft, ngood)

       ! iFt = inv(Ft)
       call dcopy(ngood*ngood, Ft, 1, iFt, 1)
       call dpotrf('u', ngood, iFt, ngood, info)
       call dpotri('u', ngood, iFt, ngood, info)


       ! det(Ft)
       detFt = determinant(Ft, ngood);

       call dsymv('u', ngood, ONE, iFt, ngood, nut, 1, ZERO, iFtnut, 1)

       if (t > t0) then
          loglh = loglh - 0.5_wp*ngood*log(2*M_PI) - 0.5_wp*log(detFt) & 
               - 0.5_wp*ddot(ngood, nut, 1, iFtnut, 1)
       endif


       ! Kt = TT*Pt*ZZ'
       call dgemm('n','t', ns, ngood, ns, ONE, TT, ns, ZZPt, ngood, ZERO, Kt, ns)

       ! At = TT*At + Kt*iFt*nut'
       call dgemv('n', ns, ngood, ONE, Kt, ns, iFtnut, 1, ZERO, gain, 1)
       call dgemv('n', ns, ns, ONE, TT, ns, At, 1, ONE, gain, 1)
       call dcopy(ns, gain, 1, At, 1)

       ! Pt = TT*Pt*TT' + RQR - Kt*iFt*Kt'
       call dgemm('n','n', ns, ns, ns, ONE, TT, ns, Pt, ns, ZERO, C, ns)
       call dsymm('r', 'u', ns, ns, ONE, Pt, ns, TT, ns, ZERO, C, ns)
       call dcopy(ns*ns, RQR, 1, Pt, 1)
       call dgemm('n', 't', ns, ns, ns, ONE, C, ns, TT, ns, ONE, Pt, ns)

       call dsymm('r', 'u', ns, ngood, ONE, iFt, ngood, Kt, ns, ZERO, KtiFt, ns)
       call dgemm('n', 't', ns, ns, ngood, NEG_ONE, KtiFt, ns, Kt, ns, ONE, Pt, ns)


       deallocate(oind)

       deallocate(Ft,iFt,iFtnut,Kt,KtiFt,ZZPt,nut,yhat)
    end do

    call mkl_free_buffers()
  end function kalman_filter_missing


  subroutine kalman_filter_missing_with_states(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0, &
    loglh, smooth_states, smooth_vars, smooth_shocks) 
    ! Evaluating the log likelihood of LGSS via the Kalman Filter.
    !  Can handle missing data entered as `NaN.'
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ny, nobs, neps, ns, t0
    real(wp), intent(in) :: y(:,:), TT(:,:), RR(:,:), QQ(:,:), DD(:), ZZ(:,:), HH(:,:)
    real(wp),intent(out) :: loglh, smooth_states(nobs,ns), smooth_shocks(nobs,neps), smooth_vars(ns,ns,nobs)

    real(wp) :: At(ns), Pt(ns,ns), RQR(ns,ns), QQRRp(neps,ns)
    real(wp) :: detFt
    integer :: t,i,j, info

    integer :: nmiss, ngood
    integer, allocatable :: oind(:)
    real(wp), allocatable :: iFtnut(:),ZZPt(:,:), Ft(:,:), iFt(:,:), Kt(:,:), KtiFt(:,:), nut(:),yhat(:), L(:,:)
    real(wp) :: ZZP0(ny,ns), gain(ns), C(ns,ns),  TTPt(ns,ns), HHmiss(ny,ny), beta_(ns,ns,nobs), eye(ns,ns)

    real(wp) :: fcst_error(ny,nobs), L_mat(ny,ns,nobs), P_mat(ns,ns,nobs), cc(ns),cc_old(ns), psi_(ns,ns)
    ! BLAS functions
    real(wp) :: ddot

    ! initialization 
    At = 0.0_wp
    call dgemm('n','t', neps, ns, neps, 1.0_wp, QQ, neps, RR, ns, 0.0_wp, QQRRp, neps)
    call dgemm('n','n', ns, ns, neps, 1.0_wp, RR, ns, QQRRp, neps, 0.0_wp, RQR, ns)

    call dlyap(TT, RQR, Pt, ns, info)

    ! Pt = TT*Pt*TT' + RQR

    call dgemm('n','n', ns, ns, ns, 1.0_wp, TT, ns, Pt, ns, 0.0_wp, TTPt, ns)
    Pt = RQR
    call dgemm('n','t', ns, ns, ns, 1.0_wp, TTPt, ns, TT, ns, 1.0_wp, Pt, ns)

    eye = 0.0_wp
    do t = 1,ns
       eye(t,t) = 1.0_wp
    end do
    fcst_error = 0.0_wp  
    
    loglh = 0.0_wp
    P_mat = 0.0_wp
    L_mat = 0.0_wp
    do t = 1, nobs
       HHmiss = 0.0_wp

       nmiss = count(isnan(y(:,t)))

       ngood = ny - nmiss
       allocate(oind(ngood),iFtnut(ngood),ZZPt(ngood, ns),Ft(ngood, ngood), &
            iFt(ngood,ngood),Kt(ns,ngood), KtiFt(ns,ngood),nut(ngood),yhat(ngood),&
            L(ngood,ns))

       j = 1;
       do i = 1,ny
          if (isnan(y(i,t))) then 
          else
             oind(j) = i
             j = j + 1;
          end if
       end do


       smooth_states(t,:) = At
       P_mat(:,:,t) = Pt

       ! yhat = ZZ*At + DD
       call dcopy(ngood, DD(oind), 1, yhat, 1)
       call dgemv('n', ngood, ns, ONE, ZZ(oind,:), ngood, At, 1, ONE, yhat, 1)

       ! nut = yt - yhat
       nut = y(oind, t) - yhat
       fcst_error(1:ngood,t) = nut

       ! Ft = ZZ*Pt*ZZ' + HH
       call dcopy(ngood*ngood, HH(oind, oind), 1, Ft, 1)

       !call dsymm('r', 'l', ngood, ns, ONE, Pt, ns, ZZ(oind,:), ngood, ZERO, ZZPt, ngood)
       call dgemm('n','n', ngood, ns, ns, ONE, ZZ(oind, :), ngood, Pt, ns, ZERO, ZZPt, ngood)
       call dgemm('n', 't', ngood, ngood, ns, ONE, ZZPt, ngood, ZZ(oind,:), ngood, ONE, Ft, ngood)

       ! iFt = inv(Ft)
       call dcopy(ngood*ngood, Ft, 1, iFt, 1)
       call dpotrf('u', ngood, iFt, ngood, info)
       call dpotri('u', ngood, iFt, ngood, info)


       ! det(Ft)
       detFt = determinant(Ft, ngood);

       call dsymv('u', ngood, ONE, iFt, ngood, nut, 1, ZERO, iFtnut, 1)

       if (t > t0) then
          loglh = loglh - 0.5_wp*ngood*log(2*M_PI) - 0.5_wp*log(detFt) & 
               - 0.5_wp*ddot(ngood, nut, 1, iFtnut, 1)
       endif

       ! Kt = TT*Pt*ZZ'
       call dgemm('n','t', ns, ngood, ns, ONE, TT, ns, ZZPt, ngood, ZERO, Kt, ns)

       ! At = TT*At + Kt*iFt*nut'
       call dgemv('n', ns, ngood, ONE, Kt, ns, iFtnut, 1, ZERO, gain, 1)
       call dgemv('n', ns, ns, ONE, TT, ns, At, 1, ONE, gain, 1)
       call dcopy(ns, gain, 1, At, 1)

       ! Pt = TT*Pt*TT' + RQR - Kt*iFt*Kt'
       call dgemm('n','n', ns, ns, ns, ONE, TT, ns, Pt, ns, ZERO, C, ns)
       call dsymm('r', 'u', ns, ns, ONE, Pt, ns, TT, ns, ZERO, C, ns)
       call dcopy(ns*ns, RQR, 1, Pt, 1)
       call dgemm('n', 't', ns, ns, ns, ONE, C, ns, TT, ns, ONE, Pt, ns)

       call dsymm('r', 'u', ns, ngood, ONE, iFt, ngood, Kt, ns, ZERO, KtiFt, ns)
       call dgemm('n', 't', ns, ns, ngood, NEG_ONE, KtiFt, ns, Kt, ns, ONE, Pt, ns)

       ! from Hess's code
       !call dgemm('t','n',ns,ngood,ngood,ONE,ZZ(oind,:),ngood,iFt,ngood,0.0_wp,L,ns)
       call dsymm('l','u',ngood,ns,ONE,iFt,ngood,ZZ(oind,:),ngood,0.0_wp,L,ngood)
       L_mat(1:ngood,:,t) = L

       beta_(:,:,t) = eye
       !call dcopy(ns*ns,eye, 1, beta_(t,:,:),1)
       call dgemm('t','n',ns,ns,ngood,-1.0_wp,L,ngood,ZZPt,ngood,ONE,beta_(:,:,t),ns)

       ! if (t == nobs-1) then
       !    open(2,file='beta_pre.txt',action='write')
       !    do j = 1,ns
       !       write(2,'(200f32.8)') beta_(j,:,t)
       !    end do
       !    close(2)

       !    open(1,file='L2.txt',action='write')
       !    do j = 1,ngood
       !       write(1,'(200f32.8)') L(j,:)
       !    end do
       !    close(1)

       !    open(1,file='ZZPt.txt',action='write')
       !    do j = 1,ngood
       !       write(1,'(200f32.8)') ZZPt(j,:)
       !    end do
       !    close(1)

       ! end if

       if (t > 1) then
          call dgemm('n','t',ns,ns,ns,1.0_wp,beta_(:,:,t-1),ns,TT,ns,ZERO,TTPt,ns)
          !call dcopy(ns*ns, TTPt, 1, beta_(:,:,t-1), 1)
          beta_(:,:,t-1) = TTPt
       end if

       ! if (t == nobs) then
       !    open(2,file='beta_post.txt',action='write')
       !    do j = 1,ns
       !       write(2,'(200f32.8)') beta_(:,j,t-1)
       !    end do
       !    close(2)
       ! end if

       deallocate(oind)

       deallocate(Ft,iFt,iFtnut,Kt,KtiFt,ZZPt,nut,yhat,L)
    end do


    ! very rudimentary smoother
    cc = 0.0_wp
    cc_old = 0.0_wp

    psi_ = 0.0_wp

    do t=nobs,1,-1

       nmiss = count(isnan(y(:,t)))

       ngood = ny - nmiss
       allocate(oind(ngood))
       j = 1;
       do i = 1,ny
          if (isnan(y(i,t))) then 
          else
             oind(j) = i
             j = j + 1;
          end if
       end do

       ! psi = beta*psi_*beta' + L'ZZ;
       call dgemm('n','t',ns,ns,ns,ONE,psi_,ns,beta_(:,:,t),ns,ZERO,TTPt,ns)
       call dgemm('n','n',ns,ns,ns,ONE,beta_(:,:,t),ns,TTPt,ns,ZERO,psi_,ns)
       call dgemm('t','n',ns,ns,ngood,ONE,L_mat(1:ngood,:,t),ngood,ZZ(oind,:),ngood,ONE,psi_,ns)

       ! c = L_t*eta_t + beta_t*c
       ! a_tT = a_t + P_t*c
       cc_old = cc
       call dgemv('n',ns,ns,ONE,beta_(:,:,t),ns,cc_old,1,ZERO,cc,1)
       call dgemv('t',ngood,ns,ONE,L_mat(1:ngood,:,t),ngood,fcst_error(1:ngood,t),1,ONE,cc,1)
       call dgemv('n',ns,ns,ONE,P_mat(:,:,t),ns,cc,1,ONE,smooth_states(t,:),1)

       call dgemv('n',neps,ns,ONE,QQRRp,neps,cc,1,ONE,smooth_shocks(t,:),1)

       smooth_vars(:,:,t) = P_mat(:,:,t)
       call dgemm('n','n',ns,ns,ns,NEG_ONE,psi_,ns,P_mat(:,:,t),ns,ZERO,TTPt,ns)
       call dgemm('n','n',ns,ns,ns,ONE,P_mat(:,:,t),ns,psi_,ns,ONE,smooth_vars,ns)
       

       deallocate(oind)

    end do

    call mkl_free_buffers()
  end subroutine kalman_filter_missing_with_states



  function chand_recursion(y, TT, RR, QQ, DD, ZZ, HH, ny, nobs, neps, ns, t0) result(loglh)
    ! Evaluating the log likelihood of a LGSS via the Chandrasekhar Recursions.
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ny, nobs, neps, ns, t0
    real(wp), intent(in) :: y(:,:), TT(:,:), RR(:,:), QQ(:,:), DD(:), ZZ(:,:), HH(:,:)
    real(wp) :: loglh

    real(wp) :: Kt(ns, ny), St(ns, ny), Mt(ny, ny), yhat(ny), nut(ny), At(ns), RQR(ns,ns)
    real(wp) :: Ft(ny, ny), iFt(ny, ny), Ft1(ny, ny), iFt1(ny, ny), MSpZp(ny, ny), P0(ns,ns)
    real(wp) :: detFt, Ft2(2,2), Kt1(ns,ny)
    integer :: t

    ! for blas / lapack
    real(wp) :: ZZP0(ny, ns), TTP0(ns,ns), gain(ns), iFtnut(ny), ZZSt(ny,ny), ZZStMt(ny,ny)
    real(wp) :: KtFt(ns, ny), TTSt(ns,ny), QQRRp(neps,ns)
    real(wp) :: ddot
    integer :: info

    ! for slicot
    real(wp) :: scale, U(ns,ns), UH(ns, ns), rcond, ferr, wr(ns), wi(ns), dwork(5*ns*ns), sepd
    integer :: iwork(ns*ns), ldwork
    double precision :: kk

    real(wp) :: tol
    logical :: converged

    loglh = 0.0_wp

    ! initialization 
    At = 0.0_wp

    call dgemm('n','t', neps, ns, neps, 1.0_wp, QQ, neps, RR, ns, 0.0_wp, QQRRp, neps)
    call dgemm('n','n', ns, ns, neps, 1.0_wp, RR, ns, QQRRp, neps, 0.0_wp, RQR, ns)

    call dlyap(TT, RQR, P0, ns, info)

    ! Pt = TT*Pt*TT' + RQR
    !call dgemm('n','n', ns, ns, ns, 1.0_wp, TT, ns, P0, ns, 0.0_wp, TTP0, ns)
    call dsymm('r', 'l', ns, ns, 1.0_wp, P0, ns, TT, ns, 0.0_wp, TTP0, ns)
    P0 = RQR
    call dgemm('n','t', ns, ns, ns, 1.0_wp, TTP0, ns, TT, ns, 1.0_wp, P0, ns)

    ! Ft = ZZ*P0*ZZ' + HH
    call dcopy(ny*ny, HH, 1, Ft, 1)
    call dsymm('r', 'l', ny, ns, ONE, P0, ns, ZZ, ny, ZERO, ZZP0, ny)
    call dgemm('n', 't', ny, ny, ns, ONE, ZZP0, ny, ZZ, ny, ONE, Ft, ny)

    ! iFt = inv(Ft)
    !call dcopy(ny*ny, Ft, 1, iFt, 1)
    iFt = 0.0_wp
    call dlacpy('u', ny, ny, Ft, ny, iFt, ny)
    call dpotrf('u', ny, iFt, ny, info)
    call dpotri('u', ny, iFt, ny, info)


    !St = TT*P0*ZZ.transpose();
    call dsymm('r', 'u', ns, ns, ONE, P0, ns, TT, ns, ZERO, TTP0, ns)
    call dgemm('n', 't', ns, ny, ns, ONE, TTP0, ns, ZZ, ny, ZERO, St, ns)


    ! Mt = -iFt
    Mt = 0.0_wp                 ! this initialization is necessary!
    call daxpy(ny*ny, NEG_ONE, iFt, 1, Mt, 1)

    ! Kt = St*iFt
    call dsymm('r', 'u', ns, ny, ONE, iFt, ny, St, ns, ZERO, Kt, ns)
    !    call dgemm('n','n',ns, ny, ny, ONE, St, ns, iFt, ny, ZERO, Kt, ns)

    tol = 1e-5
    converged = .false.


    do t = 1, nobs

       ! yhat = ZZ*At + DD
       call dcopy(ny, DD, 1, yhat, 1)
       call dgemv('n', ny, ns, ONE, ZZ, ny, At, 1, ONE, yhat, 1)

       ! nut = yt - yhat
       nut = y(:, t) - yhat

       ! det
       detFt = determinant(Ft, ny)

       call dsymv('u', ny, ONE, iFt, ny, nut, 1, ZERO, iFtnut, 1)
       if (t > t0) then
          loglh = loglh - 0.5_wp*ny*log(2*M_PI) - 0.5_wp*log(detFt) & 
               - 0.5_wp*ddot(ny, nut, 1, iFtnut, 1)
       endif

       ! At = TT*At + Kt*nut
       call dgemv('n', ns, ny, ONE, Kt, ns, nut, 1, ZERO, gain, 1)
       call dgemv('n', ns, ns, ONE, TT, ns, At, 1, ONE, gain, 1)
       call dcopy(ns, gain, 1, At, 1)

       if (converged == .false.) then 
          ! Ft1 = Ft + ZZ*St*Mt*St'ZZ';
          call dcopy(ny*ny, Ft, 1, Ft1, 1)
          call dgemm('n', 'n', ny, ny, ns, ONE, ZZ, ny, St, ns, ZERO, ZZSt, ny)
          call dsymm('r', 'u', ny, ny, ONE, Mt, ny, ZZSt, ny, ZERO, ZZStMt, ny)
          call dgemm('n', 't', ny, ny, ny, ONE, ZZStMt, ny, ZZSt, ny, ONE, Ft1, ny)
          !Ft1 = 0.5_wp*Ft1 + 0.5_wp*transpose(Ft1)

          ! iFt1 = inv(Ft1)
          call dcopy(ny*ny, Ft1, 1, iFt1, 1)
          call dpotrf('u', ny, iFt1, ny, info)
          call dpotri('u', ny, iFt1, ny, info)

          ! Kt = (Kt*Ft + TT*St*MSpZp)*iFt1;
          call dsymm('r', 'u', ns, ny, ONE, Ft, ny, Kt, ns, ZERO, KtFt, ns)
          call dgemm('n', 'n', ns, ny, ns, ONE, TT, ns, St, ns, ZERO, TTSt, ns)
          call dgemm('n', 't', ns, ny, ny, ONE, TTSt, ns, ZZStMt, ny, ONE, KtFt, ns)
          call dsymm('r', 'u', ns, ny, ONE, iFt1, ny, KtFt, ns, ZERO, Kt1, ns)


          ! St = (TT - Kt*ZZ)*St = TT*St - Kt*ZZ*St
          call dgemm('n','n', ns, ny, ny, NEG_ONE, Kt1, ns, ZZSt, ny, ONE, TTSt, ns)
          call dcopy(ny*ns, TTSt, 1, St, 1)


          ! Mt = Mt + MSpZp*iFt*MSpZp.transpose();
          call dsymm('l','u', ny, ny, ONE, iFt, ny, ZZStMt, ny, ZERO, Ft, ny)
          call dgemm('t','n', ny, ny, ny, ONE, ZZStMt, ny, Ft, ny, ONE, Mt, ny)

          if (maxval(abs(Kt1 - Kt)) < tol) then
             converged = .true.
          endif

          ! Ft = Ft1, iFt = iFt1
          call dcopy(ny*ny, Ft1, 1, Ft, 1)
          call dcopy(ny*ny, iFt1, 1, iFt, 1)

          call dcopy(ny*ns, Kt1, 1, Kt, 1)

       end if
    end do
  end function chand_recursion



  !----------------------------------------------------------------------------------------------------
  ! Helper functions/subroutines for filtering functions.
  !   - determinant
  !----------------------------------------------------------------------------------------------------
  function determinant(matrix, r) result(det)
    ! Computes the determinant of symmetric square matrix, matrix (rank r).
    integer, intent(in) :: r
    real(wp), intent(in) :: matrix(r,r)

    real(wp) :: det

    integer :: info, i, piv(r)
    real(wp) :: matrix_copy(r, r)

    call dcopy(r*r, matrix, 1, matrix_copy, 1)
    det = 0.0_wp

    call dpotrf('u', r, matrix_copy, r, info)

    if (info .ne. 0) then
       !write(*,'(a,i4)') 'In determinant(), dgetrf returned error code ', info
       det = -10000.0_wp
       return
    end if

    det = ONE

    do i = 1, r

       if (.true.) then !(piv(i) .ne. i) then
          det = det * matrix_copy(i, i) * matrix_copy(i,i)
       else
          det = det * matrix_copy(i, i)
       end if

    end do
  end function determinant


  subroutine dlyap(TT, RQR, P0, ns, info)
    ! Computes the solution to the discrete Lyapunov equation,
    !      P0 = TT*P0*TT' + RQR
    ! where (inputs) TT, RQR and (output) P0 are ns x ns (real) matrices. 
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ns
    real(wp), intent(in) :: TT(ns,ns), RQR(ns,ns)

    integer, intent(out) :: info
    real(wp), intent(out) :: P0(ns,ns)

    ! for slicot
    real(wp) :: scale, U(ns,ns), UH(ns, ns), rcond, ferr, wr(ns), wi(ns), dwork(14*ns*ns*ns), sepd
    integer :: iwork(ns*ns), ldwork

    integer :: t

    UH = TT
    P0 = -1.0_wp*RQR

    call sb03md('D','X', 'N', 'T', ns, UH, ns, U, ns, P0, ns, &
         scale, sepd, ferr, wr, wi, iwork, dwork, 14*ns*ns*ns, info)

    !if (ferr > 0.000001_wp) call dlyap_symm(TT, RQR, P0, ns, info)
    if (info .ne. 0) then
       print*,'SB03MD failed. (info = ', info, ')'
       P0 = 0.0_wp
       info = 1
       do t = 1,ns
          P0(t,t)=1.0_wp
       end do

       return
    else

!       P0 = 0.5_wp*P0 + 0.5_wp*transpose(P0)
       info = 0

    end if

  end subroutine dlyap





  !----------------------------------------------------------------------------------------------------
  !
  ! State space system reduction.
  !
  !----------------------------------------------------------------------------------------------------
  
  subroutine reduce_system(TT,RR,ZZ,TTred,RRred,ZZred,ny,ns,neps,ns1,check_dim)

    integer, intent(in) :: ny, ns, neps
    integer, intent(out) :: ns1

    real(wp), intent(in) :: TT(ns,ns), RR(ns,neps), ZZ(ny,ns)
    real(wp), intent(out) :: TTred(:,:), RRred(:,:), ZZred(:,:)

    logical, intent(in) :: check_dim 

    ! for Schur Decomposition
    integer :: nmin, info, nwork
    real(wp) :: U(ns,ns), Q(ns,ns)
    logical :: bwork(ns)
    real(wp) :: wr(ns), wi(ns), work(3*ns)

    real(wp) :: RRtemp(ns,neps)
    integer :: i, j

    real(wp), allocatable :: temp(:,:), ZZQ(:,:)
    integer :: nconteps, conteps(neps)


    U = TT

    call dgees('v','s', selctg, ns, U, ns, nmin, wr, wi, Q, ns, work, 3*ns, bwork, info)

    call dgemm('t','n',ns,neps,ns,1.0_wp, Q, ns, RR, ns, 0.0_wp, RRtemp, ns)

    nconteps = count(sum(abs(RRtemp(nmin+1:ns,:)),dim=1)>really_small)

    ns1 = nmin + nconteps
    if (check_dim == .true.) return

    j = 1
    do i = 1,neps
       if (sum(abs(RRtemp(nmin+1:ns,i))) > really_small) then
          conteps(j) = i
          j = j + 1
       end if
    end do

    TTred = 0.0_wp
    RRred = 0.0_wp
    ZZred = 0.0_wp

    if (nconteps == 0) then

       TTred = U(1:ns1,1:ns1)
       RRred = RRtemp(1:ns,:);
       call dgemm('n','n',ny, ns1, ns, 1.0_wp, ZZ, ny, Q(:,1:ns1), ns, 0.0_wp, ZZred, ny)

    else

       allocate(temp(ns,ns1), ZZQ(ny,ns))

       TTred(1:nmin,1:nmin) = U(1:nmin,1:nmin)
       call dgemm('n','n', nmin, nconteps, ns-nmin, 1.0_wp, U(1:nmin,nmin+1:ns), nmin, RRtemp(nmin+1:ns,conteps(1:nconteps)), ns-nmin, 0.0_wp, TTred(1:nmin,nmin+1:ns1), nmin)

       RRred(1:nmin,:) = RRtemp(1:nmin,:)
       j = 1;
       do i = nmin+1,ns1
          RRred(i,conteps(j)) = 1.0_wp
          j = j + 1
       end do

       temp = 0.0_wp
       do i = 1,nmin
          temp(i,i) = 1.0_wp
       end do
       temp(nmin+1:ns,nmin+1:ns1) = RRtemp(nmin+1:ns,conteps(1:nconteps))

       call dgemm('n','n', ny, ns, ns, 1.0_wp, ZZ, ny, Q, ns, 0.0_wp, ZZQ, ny)
       call dgemm('n','n', ny, ns1, ns, 1.0_wp, ZZQ, ny, temp, ns, 0.0_wp, ZZred, ny)

       deallocate(temp,ZZQ)

    end if

  end subroutine reduce_system

  logical function selctg(wi, wr)

    real(wp) :: wi, wr

    if (sqrt(wi**2 + wr**2) > 1e-15) then 
       selctg = .true.
    else
       selctg = .false.
    end if

  end function selctg


  !----------------------------------------------------------------------------------------------------
  !
  ! Gradient of the log likelihood via the Chandrasekhar recursions and assorted helper functions.
  !
  !----------------------------------------------------------------------------------------------------

  subroutine chand_recursion_derivative(y, TT, RR, QQ, DD, ZZ, HH, DTT, DRR, DQQ, DDD, DZZ, DHH, & 
       ny, nobs, neps, ns, t0, neffpara, loglh, dloglh, info_mat)
    ! Evaluates the log likelihood and the gradient of the log likelihood of a LGSS.
    !--------------------------------------------------------------------------------
    integer, intent(in) :: ny, nobs, neps, ns, t0, neffpara
    real(wp), intent(in) :: y(:,:), TT(:,:), RR(:,:), QQ(:,:), DD(:), ZZ(:,:), HH(:,:)
    real(wp), intent(in) :: DTT(:,:), DRR(:,:), DQQ(:,:), DDD(:,:), DZZ(:,:), DHH(:,:)
    real(wp) :: loglh


    real(wp) :: Kt(ns, ny), St(ns, ny), Mt(ny, ny), yhat(ny), nut(ny), At(ns), RQR(ns,ns)
    real(wp) :: Ft(ny, ny), iFt(ny, ny), Ft1(ny, ny), iFt1(ny, ny), MSpZp(ny, ny), P0(ns,ns)
    real(wp) :: detFt, Ft2(2,2)
    integer :: t

    ! for blas / lapack
    real(wp) :: ZZP0(ny, ns), TTP0(ns,ns), gain(ns), iFtnut(ny), ZZSt(ny,ny), ZZStMt(ny,ny)
    real(wp) :: KtFt(ns, ny), TTSt(ns,ny), QQRRp(neps,ns)
    real(wp) :: ddot
    integer :: info

    real(wp) :: Ins(ns,ns), Iny(ny,ny), Kss(ns**2, ns**2), Kys(ny*ns, ny*ns), Ksy(ny*ns, ny*ns), Kyy(ny**2, ny**2)
    real(wp) :: Is2(ns**2,ns**2), Iy2(ny**2,ny**2), Nns(ns**2, ns**2), Nny(ny**2,ny**2)

    real(wp) :: TTkrTT(ns**2,ns**2), KrTTPt(ns**2, ns**2), KrTTPt2(ns**2, ns**2)
    real(wp) :: KrRQI(ns**2, ns*neps), NnsKrRQI(ns**2, ns*neps), DRQRp(ns**2, neffpara), RRKrRRDQQ(ns**2, neffpara)
    real(wp) :: RRKrRR(ns**2, neps**2), ZZkrZZ(ny**2, ns**2), ZZP0krDZZ(ny**2, neffpara), DF(ny**2,neffpara)

    real(wp) :: DiFt(ny**2, neffpara), iFtkriFt(ny**2,ny**2), DMt(ny**2, neffpara), DSt(ny*ns, neffpara), DSt2(ny*ns,neffpara)
    real(wp) :: ZZkrTT(ny*ns,ns*ns), DKt(ny*ns, neffpara), DKt1(ny*ns, neffpara), nutkrnut(1,ny**2)

    real(wp) :: DAt(ns,neffpara), DNU(ny,neffpara), DAt1(ns,neffpara), DZSM(ny*ny,neffpara), MkrZZ(ny*ny,ny*ns), MSp(ny,ns)
    real(wp) :: DM2(ny**2,neffpara), DM3(ny**2,neffpara), KDS(ny*ns, neffpara), DZSMS(ny*ns,neffpara)

    real(wp) :: DiFt1(ny**2,neffpara), DFt1(ny**2,neffpara), ZZStMtSp(ny,ns), temp(ny*ny,neffpara), Kt1(ns,ny)
    real(wp) :: iFt1krKt(ny*ns, ny**2), iFt1Ft(ny,ny), iFt1krTT(ny*ns,ny*ns), TTStMtSpZp(ns,ny), DSt1(ny*ns,neffpara)

    real(wp) :: KtZZ(ns,ns), info_mat(neffpara,neffpara), temp2(ny,neffpara)

    real(wp) :: tol
    logical :: converged

    ! just for testing
    real(wp) :: ZZP0krIny(ny*ny,ns*ny), KDZZ(ny*ns,neffpara)

    real(wp), intent(out) :: dloglh(neffpara)

    integer :: ipiv(ns**2)
    integer :: i

    ! construct Ins, Iny
    Ins = 0.0_wp
    Is2 = 0.0_wp
    Iny = 0.0_wp
    Iy2 = 0.0_wp

    do i = 1, ns 
       Ins(i,i) = 1.0_wp
    end do

    do i = 1,ns**2
       Is2(i,i) = 1.0_wp
    end do

    do i = 1,ny**2
       Iy2(i,i) = 1.0_wp
    end do

    do i = 1, ny 
       Iny(i,i) = 1.0_wp
    end do

    Kss = commutation_matrix(ns,ns)
    Ksy = commutation_matrix(ns,ny)
    Kyy = commutation_matrix(ny,ny)
    Kys = transpose(Ksy)

    KDZZ = matmul(Kys, DZZ)
    Nns = Kss + Is2
    Nny = Kyy + Iy2

    loglh = 0.0_wp
    dloglh = 0.0_wp
    info_mat = 0.0_wp
    ! initialization 
    At = 0.0_wp


    call dgemm('n','t', neps, ns, neps, 1.0_wp, QQ, neps, RR, ns, 0.0_wp, QQRRp, neps)
    call dgemm('n','n', ns, ns, neps, 1.0_wp, RR, ns, QQRRp, neps, 0.0_wp, RQR, ns)

    call dlyap(TT, RQR, P0, ns, info)

    ! Pt = TT*Pt*TT' + RQR
    !call dgemm('n','n', ns, ns, ns, 1.0_wp, TT, ns, P0, ns, 0.0_wp, TTP0, ns)
    call dsymm('r', 'l', ns, ns, 1.0_wp, P0, ns, TT, ns, 0.0_wp, TTP0, ns)
    P0 = RQR
    call dgemm('n','t', ns, ns, ns, 1.0_wp, TTP0, ns, TT, ns, 1.0_wp, P0, ns)

    ! Ft = ZZ*P0*ZZ' + HH
    call dcopy(ny*ny, HH, 1, Ft, 1)
    call dsymm('r', 'l', ny, ns, ONE, P0, ns, ZZ, ny, ZERO, ZZP0, ny)
    call dgemm('n', 't', ny, ny, ns, ONE, ZZP0, ny, ZZ, ny, ONE, Ft, ny)

    ! iFt = inv(Ft)
    iFt = 0.0_wp
    call dlacpy('u', ny, ny, Ft, ny, iFt, ny)
    call dpotrf('u', ny, iFt, ny, info)
    call dpotri('u', ny, iFt, ny, info)

    call utri2dens(iFt, ny)

    !St = TT*P0*ZZ.transpose();
    call dsymm('r', 'u', ns, ns, ONE, P0, ns, TT, ns, ZERO, TTP0, ns)
    call dgemm('n', 't', ns, ny, ns, ONE, TTP0, ns, ZZ, ny, ZERO, St, ns)

    ! Mt = -iFt
    Mt = 0.0_wp                 ! this initialization is necessary!
    call daxpy(ny*ny, NEG_ONE, iFt, 1, Mt, 1)

    ! Kt = St*iFt
    call dsymm('r', 'u', ns, ny, ONE, iFt, ny, St, ns, ZERO, Kt, ns)


    call Kronecker(ns, ns, ns, ns, ns**2, ns**2, 0.0_wp, 1.0_wp, TT, TT, TTkrTT)

    !----------fajdskflas
    !call dgemm('n','n',ns,ns,ns,1.0_wp,TT,ns,P0,ns,0.0_wp,TTP0,ns)
    !-----------fjdakslf
    call Kronecker(ns, ns, ns, ns, ns**2, ns**2, 0.0_wp, 1.0_wp, TTP0, Ins, KrTTPt)
    call dgemm('n','n', ns**2, ns**2, ns**2, 1.0_wp, Nns, ns**2, KrTTPt, ns**2, 0.0_wp, KrTTPt2, ns**2)

    call Kronecker(ns, neps, ns, ns, ns**2, ns*neps, 0.0_wp, 1.0_wp, transpose(QQRRp), Ins, KrRQI)
    call dgemm('n','n', ns**2, ns*neps, ns**2, 1.0_wp, Nns, ns**2, KrRQI, ns**2, 0.0_wp, NnsKrRQI, ns**2)

    call Kronecker(ns, neps, ns, neps, ns**2, neps**2, 0.0_wp, 1.0_wp, RR, RR, RRKrRR)

    call dgemm('n','n',ns**2, neffpara, neps**2, 1.0_wp, RRKrRR, ns**2, DQQ, neps**2, 0.0_wp, RRKrRRDQQ, ns**2)
    call dgemm('n','n',ns**2, neffpara, ns*neps, 1.0_wp, NnsKrRQI, ns**2, DRR, ns*neps, 1.0_wp, RRkrRRDQQ, ns**2)


    call dgemm('n','n',ns**2, neffpara, ns**2, 1.0_wp, KrTTPt2, ns**2, DTT, ns**2, 1.0_wp, RRkrRRDQQ, ns**2)

    !    call daxpy(ns**2, -1.0_wp, TTkrTT, 1, Is2, 1)

    Is2 = Is2 - TTkrTT

    call dgesv(ns**2, neffpara, Is2, ns**2, ipiv, RRkrRRDQQ, ns**2, info)

    ! at this point DP = RRkrRRDQQ
    ! DFt 
    call dcopy(ny**2*neffpara, DHH, 1, DF, 1)
    call Kronecker(ny, ns, ny, ns, ny**2, ns**2, 0.0_wp, 1.0_wp, ZZ, ZZ, ZZkrZZ)

    ! Instead of krlm!!!!
    !call krlm(ny, ns, ns**2, neffpara, ZZP0, DZZ, ZZP0krDZZ)
    call Kronecker(ny, ns, ny, ny, ny**2, ns*ny, 0.0_wp, 1.0_wp, ZZP0, Iny, ZZP0krIny)
    call dgemm('n','n', ny*ny, neffpara, ns*ny, 1.0_wp, ZZP0krIny, ny*ny, DZZ, ny*ns, 0.0_wp, ZZP0krDZZ, ny*ny)

    call dgemm('n','n', ny**2, neffpara, ny**2, 1.0_wp, Nny, ny**2, ZZP0krDZZ, ny**2, 1.0_wp, DF, ny**2)
    call dgemm('n','n', ny**2, neffpara, ns**2, 1.0_wp, ZZkrZZ, ny**2, RRkrRRDQQ, ns**2, 1.0_wp, DF, ny**2)

    ! DiFt
    call Kronecker(ny, ny, ny, ny, ny**2, ny**2, 0.0_wp, 1.0_wp, iFt, iFt, iFtkriFt)
    !call dsymm('l','u', ny**2, neffpara, -1.0_wp, iFtkriFt, ny**2, DF, ny**2, 0.0_wp, DiFt, ny**2)
    call dgemm('n','n', ny**2, neffpara, ny**2, -1.0_wp, iFtkriFt, ny**2, DF, ny**2, 0.0_wp, DiFt, ny**2)

    ! DMt
    DMt = -DiFt

    ! DSt
    call Kronecker(ny,ns,ns,ns, ny*ns, ns*ns, 0.0_wp, 1.0_wp, ZZ, TT, ZZkrTT)
    call krlm(ny, ns, ns**2, neffpara, ZZP0, DTT, DSt)
    call krrm(ns, ns, ny*ns, neffpara, TTP0, KDZZ, DSt2)
    call daxpy(ny*ns*neffpara, 1.0_wp, DSt2, 1, DSt, 1)

    call dgemm('n','n', ny*ns, neffpara, ns**2, 1.0_wp, ZZkrTT, ny*ns, RRkrRRDQQ, ns**2, 1.0_wp, DSt, ny*ns)

    ! DKt
    DKt = 0.0_wp
    DKt1 = 0.0_wp

    call krlm(ny, ny, ny*ns, neffpara, iFt, DSt, DKt)
    call krrm(ns, ny, ny**2, neffpara, St, DiFt, DKt1)
    call daxpy(ny*ns*neffpara, 1.0_wp, DKt1, 1, DKt, 1)

    tol = 1e-5
    converged = .false.

    DAt = 0.0_wp

    do t = 1, nobs

       ! yhat = ZZ*At + DD
       call dcopy(ny, DD, 1, yhat, 1)
       call dgemv('n', ny, ns, ONE, ZZ, ny, At, 1, ONE, yhat, 1)

       ! nut = yt - yhat
       nut = y(:, t) - yhat

       ! det
       detFt = determinant(Ft, ny)

       call dsymv('u', ny, ONE, iFt, ny, nut, 1, ZERO, iFtnut, 1)

       ! DNU
       DNU = 0.0_wp
       call krlm(1, ns, ns*ny, neffpara, At, DZZ, DNU)
       call daxpy(ny*neffpara, 1.0_wp, DDD, 1, DNU, 1)
       call dgemm('n','n', ny, neffpara, ns, -1.0_wp, ZZ, ny, DAt, ns, -1.0_wp, DNU, ny)

       call Kronecker(1, ny, 1, ny, 1, ny**2, 0.0_wp, 1.0_wp, nut, nut, nutkrnut)

       if (t > t0) then
          loglh = loglh - 0.5_wp*ny*log(2*M_PI) - 0.5_wp*log(detFt) & 
               - 0.5_wp*ddot(ny, nut, 1, iFtnut, 1)

          ! gradient
          call dgemm('n','n', 1, neffpara, ny**2, -0.5_wp, iFt, 1, DF, ny**2, 1.0_wp, dloglh, 1)
          call dgemv('t', ny, neffpara, -1.0_wp, DNU, ny,  iFtnut, 1, 1.0_wp, dloglh, 1)
          call dgemv('t', ny**2, neffpara, -0.5_wp, DiFt, ny**2, nutkrnut, 1, 1.0_wp, dloglh, 1)
          !          print*,dloglh
          ! information matrix
          call Kronecker(ny, ny, ny, ny, ny**2, ny**2, 0.0_wp, 1.0_wp, iFt1, iFt1, iFtkriFt)
          call dgemm('n','n',ny**2, neffpara, ny**2, 1.0_wp, iFtkriFt, ny**2, DF, ny**2, 0.0_wp, DiFt1,ny**2)
          call dgemm('t','n',neffpara, neffpara, ny**2, 0.5_wp, DF, ny**2, DiFt1, ny**2, 1.0_wp, info_mat, neffpara)

          call dgemm('n','n', ny, neffpara, ny, 1.0_wp, iFt, ny, DNU, ny, 0.0_wp, temp2, ny)
          call dgemm('t','n', neffpara, neffpara, ny, 1.0_wp, DNU, ny, temp2, ny, 1.0_wp, info_mat, neffpara)

       endif

       ! At = TT*At + Kt*nut
       call dgemv('n', ns, ny, ONE, Kt, ns, nut, 1, ZERO, gain, 1)
       call dgemv('n', ns, ns, ONE, TT, ns, At, 1, ONE, gain, 1)

       ! At this point, At1 = gain, At is old At
       call krlm(1, ns, ns**2, neffpara, At, DTT, DAt1)
       call dgemm('n','n', ns, neffpara, ns, 1.0_wp, TT, ns, DAt, ns, 1.0_wp, DAt1, ns)
       call krlm(1, ny, ny*ns, neffpara, nut, DKt, DAt)
       call dgemm('n', 'n', ns, neffpara, ny, 1.0_wp, Kt, ns, DNU, ny, 1.0_wp, DAt, ns)
       call daxpy(ns*neffpara, 1.0_wp, DAt1, 1, DAt, 1)
       
       ! Now At1 = At
       call dcopy(ns, gain, 1, At, 1)

       if (converged == .false.) then 
          
          ! Ft1 = Ft + ZZ*St*Mt*St'ZZ';
          call dcopy(ny*ny, Ft, 1, Ft1, 1)
          call dgemm('n', 'n', ny, ny, ns, ONE, ZZ, ny, St, ns, ZERO, ZZSt, ny)
          call dsymm('r', 'u', ny, ny, ONE, Mt, ny, ZZSt, ny, ZERO, ZZStMt, ny)
          call dgemm('n', 't', ny, ny, ny, ONE, ZZStMt, ny, ZZSt, ny, ONE, Ft1, ny)
          !Ft1 = 0.5_wp*Ft1 + 0.5_wp*transpose(Ft1)

          ! DZSM
          !-------------------------------------------------------------------------------------------------
          call Kronecker(ny, ny, ny, ns, ny*ny, ny*ns, 0.0_wp, 1.0_wp, transpose(Mt), ZZ, MkrZZ)
          call dgemm('t','t', ny, ns, ny, 1.0_wp, Mt, ny, St, ns, 0.0_wp, MSp, ny)
          call krlm(ny, ns, ny*ns, neffpara, MSp, DZZ, DZSM)
          call krrm(ny, ny, ny**2, neffpara, ZZSt, DMt, DM3)
          call daxpy(ny**2*neffpara, 1.0_wp, DM3, 1, DZSM, 1)
          call dgemm('n','n', ny**2, neffpara, ns*ny, 1.0_wp, MkrZZ, ny*ny, DSt, ny*ns, 1.0_wp, DZSM, ny**2)

          ! DZSMS'
          call dgemm('n','n',ny*ns, neffpara, ny*ns, 1.0_wp, Ksy, ny*ns, DSt, ny*ns, 0.0_wp, KDS, ny*ns)
          call krrm(ny,ny,ny*ns,neffpara, ZZStMt, KDS, DZSMS)
          call krlm(ns,ny,ny**2,neffpara, St, DZSM, KDS)
          call daxpy(ny*ns*neffpara, 1.0_wp, KDS, 1, DZSMS, 1)
          call dgemm('n','t', ny, ns, ny, 1.0_wp, ZZStMt, ny, St, ns, 0.0_wp, ZZStMtSp, ny)
          !---------------------------------------------------------------------------------------------------

          ! iFt1 = inv(Ft1)
!          call dcopy(ny*ny, Ft1, 1, iFt1, 1)
          iFt1 = 0.0_wp
          call dlacpy('u', ny, ny, Ft1, ny, iFt1, ny)

          call dpotrf('u', ny, iFt1, ny, info)
          call dpotri('u', ny, iFt1, ny, info)
          call utri2dens(iFt1, ny)

          ! DFt1
          !----------------------------------------------------------------------------------------------------
          call krlm(ny, ns, ny*ns, neffpara, ZZ, DZSMS, DFt1)
          call krrm(ny, ns, ny*ns, neffpara, ZZStMtSp, KDZZ, temp)
          call daxpy(ny**2*neffpara, 1.0_wp, temp, 1, DFt1, 1)
          call daxpy(ny**2*neffpara, 1.0_wp, DF, 1, DFt1, 1)
          !-----------------------------------------------------------------------------------------------------
          
          ! DiFt1 
          !-----------------------------------------------------------------------------------------------------
          call Kronecker(ny, ny, ny, ny, ny**2, ny**2, 0.0_wp, 1.0_wp, iFt1, iFt1, iFtkriFt)
          !call dsymm('l','u', ny**2, neffpara, -1.0_wp, iFtkriFt, ny**2, DF, ny**2, 0.0_wp, DiFt, ny**2)
          call dgemm('n','n', ny**2, neffpara, ny**2, -1.0_wp, iFtkriFt, ny**2, DFt1, ny**2, 0.0_wp, DiFt1, ny**2)
          !-----------------------------------------------------------------------------------------------------
          

          ! Kt = (Kt*Ft + TT*St*MSpZp)*iFt1;
          call dsymm('r', 'u', ns, ny, ONE, Ft, ny, Kt, ns, ZERO, KtFt, ns)
          call dgemm('n', 'n', ns, ny, ns, ONE, TT, ns, St, ns, ZERO, TTSt, ns)


          call krrm(ns, ny, ny**2, neffpara, KtFt, DiFt1, DKt1)
          
          call dgemm('n', 't', ns, ny, ny, ONE, TTSt, ns, ZZStMt, ny, ONE, KtFt, ns)
          call dsymm('r', 'u', ns, ny, ONE, iFt1, ny, KtFt, ns, ZERO, Kt1, ns)

          ! DKt
          !-------------------------------------------------------------------------------------------------------
          call Kronecker(ny, ny, ns, ny, ns*ny, ny*ny, 0.0_wp, 1.0_wp, iFt1, Kt, iFt1krKt)
          call dgemm('n','n', ns*ny, neffpara, ny**2, 1.0_wp, iFt1krKt, ny*ns, DF, ny**2, 1.0_wp, DKt1, ny*ns)
          
          call dgemm('n','n', ny, ny, ny, 1.0_wp, iFt1, ny, Ft, ny, 0.0_wp, iFt1Ft, ny)
          call krlm(ny,ny,ny*ns,neffpara, iFt1Ft, DKt, KDS)
          call daxpy(ny*ns*neffpara, 1.0_wp, KDS, 1, DKt1, 1)
           

          call Kronecker(ny, ny, ns, ns, ns*ny, ns*ny, 0.0_wp, 1.0_wp, iFt1, TT, iFt1krTT)
          call dgemm('n','n', ny*ns, neffpara, ny*ns, 1.0_wp, Kys, ny*ns, DZSMS, ny*ns, 0.0_wp, KDS, ny*ns)
          call dgemm('n','n',ny*ns, neffpara, ny*ns, 1.0_wp, iFt1krTT, ny*ns, KDS, ny*ns, 1.0_wp, DKt1, ny*ns)
           
          call dgemm('n','t', ns, ny, ns, 1.0_wp, TT, ns, ZZStMtSp, ny, 0.0_wp, TTStMtSpZp, ns)
          call krrm(ns, ny, ny*ny, neffpara, TTStMtSpZp, DiFt1, KDS)
          call daxpy(ny*ns*neffpara, 1.0_wp, KDS, 1, DKt1, 1)
          call dgemm('n','n', ny, ns, ny, 1.0_wp, iFt1, ny, ZZStMtSp, ny, 0.0_wp, MSp, ny)
          call krlm(ny, ns, ns**2, neffpara, MSp, DTT, KDS)
          call daxpy(ny*ns*neffpara, 1.0_wp, KDS, 1, DKt1, 1)
          !-------------------------------------------------------------------------------------------------------

          ! St = (TT - Kt*ZZ)*St = TT*St - Kt*ZZ*St
          call dgemm('n','n', ns, ny, ny, NEG_ONE, Kt1, ns, ZZSt, ny, ONE, TTSt, ns)

          !-------------------------------------------------------------------------------------------------------
          call krrm(ns, ns, ny*ns, neffpara, TT, DSt, DSt1)
          call krlm(ny, ns, ns*ns, neffpara, transpose(St), DTT, KDS)
          call daxpy(ny*ns*neffpara, 1.0_wp, KDS, 1, DSt1, 1)

          call Kronecker(ny, ns, ns, ny, ny*ns, ny*ns, 0.0_wp, 1.0_wp, transpose(St), Kt1, iFt1krTT)
          call dgemm('n','n',ny*ns,neffpara,ny*ns, -1.0_wp, iFt1krTT, ny*ns, DZZ, ny*ns, 1.0_wp, DSt1, ny*ns)

          call krlm(ny, ny, ny*ns, neffpara, transpose(ZZSt), DKt1, KDS)
          call daxpy(ns*ny*neffpara, -1.0_wp, KDS, 1, DSt1, 1)

          ! KtZZ
          call dgemm('n','n', ns, ns, ny, 1.0_wp, Kt1, ns, ZZ, ny, 0.0_wp, KtZZ, ns)
          call krrm(ns, ns, ns*ny, neffpara, KtZZ, DSt, KDS)
          call daxpy(ns*ny*neffpara, -1.0_wp, KDS, 1, DSt1, 1)

          !-------------------------------------------------------------------------------------------------------
          call dcopy(ny*ns, TTSt, 1, St, 1)


          ! Mt = Mt + MSpZp*iFt*MSpZp.transpose();
          call dsymm('l','u', ny, ny, ONE, iFt, ny, ZZStMt, ny, ZERO, Ft, ny)
          call dgemm('t','n', ny, ny, ny, ONE, ZZStMt, ny, Ft, ny, ONE, Mt, ny)
          !-------------------------------------------------------------------------------------------------------
          call Kronecker(ny, ny, ny, ny, ny**2, ny**2, 0.0_wp, 1.0_wp, transpose(ZZStMt), transpose(ZZStMt), iFtkriFt)
          call dgemm('n','n', ny**2, neffpara, ny**2, 1.0_wp, iFtkriFt, ny**2, DiFt, ny**2, 1.0_wp, DMt, ny**2)
          call krrm(ny*ny, ny*ny, ny**2, neffpara, transpose(Ft), DZSM, DM3)
          call dgemm('n','n', ny**2, neffpara, ny**2, 1.0_wp, Nny, ny**2, DM3, ny**2, 1.0_wp, DMt, ny*ny)


          !-------------------------------------------------------------------------------------------------------
          if (maxval(abs(Kt1 - Kt)) < tol) then
             converged = .true.
          endif

          ! Ft = Ft1, iFt = iFt1
          call dcopy(ny*ny, Ft1, 1, Ft, 1)
          call dcopy(ny*ny, iFt1, 1, iFt, 1)

          call dcopy(ny*ny*neffpara, DFt1, 1, DF, 1)
          call dcopy(ny*ny*neffpara, DiFt1, 1, DiFt, 1)

          call dcopy(ny*ns, Kt1, 1, Kt, 1)
          call dcopy(ny*ns*neffpara, DKt1, 1, DKt, 1)

          call dcopy(ny*ns*neffpara, DSt1, 1, DSt, 1)

       end if

    end do

  end subroutine chand_recursion_derivative

  subroutine Kronecker(ma, na, mb, nb, mc, nc, alpha, beta, A, B, C)

    integer, intent(in) :: ma, na, mb, nb, mc, nc
    real(wp), intent(in) :: alpha, beta
    real(wp), dimension(ma,na), intent(in) :: A
    real(wp), dimension(mb,nb), intent(in) :: B
    real(wp), dimension(mc,nc), intent(inout) :: C

    integer :: i, j, i1, j1  

    do j = 1, na
       do i = 1, ma
          i1 = (i-1)*mb + 1
          j1 = (j-1)*nb + 1 
          C(i1:i1+mb-1,j1:j1+nb-1) = alpha*C(i1:i1+mb-1,j1:j1+nb-1) + beta*A(i,j)*B 
       end do
    end do

  end subroutine Kronecker

!!$  subroutine krlm_sparse(ar, ac, xtot, neffpara, A, DXval, DXrow, DXcol, nnz, res2)
!!$    ! Does the multiplication kron(A,I)*DX
!!$    !  We know that 
!!$    !    A -- ar x ac matrix 
!!$    !    DX -- xtot x neffpara
!!$    !  We can deduce that
!!$    !    X -- xr x xc
!!$    !      xc = ac
!!$    !      xr = xtot / xc
!!$    !    I -- xr x xr
!!$    !  So the resulting matrix will be 
!!$    !     res -- xr*ar x neffpara
!!$
!!$    ! because we start from vec(dX*A')
!!$    integer, intent(in) :: ar, ac, xtot, neffpara, nnz
!!$
!!$    real(wp), intent(in) :: A(ar, ac), DXval(nnz)
!!$    integer, intent(in) :: DXrow(nnz), DXcol(nnz)
!!$
!!$    real(wp) :: res(xtot*neffpara/ac, ar)
!!$
!!$    integer :: rowtran(nnz), coltran(nnz)
!!$    real(wp), intent(out) :: res2(xtot/ac*ar, neffpara)
!!$    real(wp) :: DXt(neffpara, xtot), At(ac,ar)
!!$
!!$    integer :: i, rowi, lead_int, coli
!!$
!!$    ! transpose DX
!!$    rowtran = DXcol
!!$    coltran = DXrow
!!$    
!!$    ! now reshape
!!$
!!$
!!$    lead_ind = xtot*neffpara/ac
!!$    do i = 1, nnz
!!$       rowi = rowtran(i)
!!$       coli = coltran(i)
!!$       
!!$       newcol = fmod(rowi*coli, lead_int)+1
!!$       rowtran(i) = neffpara*(coli - 1) + rowtran(i)
!!$       coltran(i) = newcol
!!$       
!!$    end do
!!$    !DXt = transpose(DX)
!!$    call mkl_domatcopy('c','t',ar ,ac, 1.0_wp, At, ar, At, ac)
!!$    
!!$!    call dgemm('N', 'T', xtot*neffpara/ac, ar, ac, 1.0_wp, DXt, (xtot*neffpara/ac), A, ar, 0.0_wp, res, xtot*neffpara/ac)
!!$    call mkl_dcoomm('n', )
!!$
!!$
!!$    res2 = transpose(reshape(res, (/neffpara, xtot/ac*ar/)))
!!$
!!$  end subroutine krlm_sparse

  subroutine krlm(ar, ac, xtot, neffpara, A, DX, res2)
    ! Does the multiplication kron(A,I)*DX
    !  We know that 
    !    A -- ar x ac matrix 
    !    DX -- xtot x neffpara
    !  We can deduce that
    !    X -- xr x xc
    !      xc = ac
    !      xr = xtot / xc
    !    I -- xr x xr
    !  So the resulting matrix will be 
    !     res -- xr*ar x neffpara

    ! because we start from vec(dX*A')
    integer, intent(in) :: ar, ac, xtot, neffpara

    real(wp), intent(in) :: A(ar, ac), DX(xtot, neffpara)

    real(wp) :: res(xtot*neffpara/ac, ar)

    real(wp), intent(out) :: res2(xtot/ac*ar, neffpara)
    real(wp) :: DXt(neffpara, xtot) 

    integer :: i

    !DXt = transpose(DX)

    call mkl_domatcopy('c','t',xtot,neffpara,1.0_wp,DX, xtot, DXt, neffpara)
    call dgemm('N', 'T', xtot*neffpara/ac, ar, ac, 1.0_wp, DXt, (xtot*neffpara/ac), A, ar, 0.0_wp, res, xtot*neffpara/ac)
    !    print*,'hi'
    !    call mkl_dimatcopy('c','t', neffpara, xtot/ac*ar, res, 1.0_wp, neffpara, xtot/ac*ar)
    res2 = transpose(reshape(res, (/neffpara, xtot/ac*ar/)))

  end subroutine krlm

  subroutine krlm_old(ar, ac, xtot, neffpara, A, DX, res2)
    ! Does the multiplication kron(A,I)*DX
    !  We know that 
    !    A -- ar x ac matrix 
    !    DX -- xtot x neffpara
    !  We can deduce that
    !    X -- xr x xc
    !      xc = ac
    !      xr = xtot / xc
    !    I -- xr x xr
    !  So the resulting matrix will be 
    !     res -- xr*ar x neffpara

    ! because we start from vec(dX*A')
    integer, intent(in) :: ar, ac, xtot, neffpara

    real(wp), intent(in) :: A(ar, ac), DX(xtot, neffpara)

    real(wp) :: res(ar*xtot/ac, xtot), eye(xtot/ac, xtot/ac)

    real(wp), intent(out) :: res2(xtot/ac*ar, neffpara)
    real(wp) :: DXt(neffpara, xtot) 

    integer :: i

    eye = 0.0_wp

    do i = 1,xtot/ac
       eye(i,i) = 1.0_wp
    end do

    call Kronecker(ar, ac, xtot/ac, xtot/ac, ar*xtot/ac, ac*xtot/ac, 0.0_wp, 1.0_wp, A, eye, res)

    call dgemm('n','n', xtot/ac*ar, neffpara, xtot, 1.0_wp, res, ar*xtot/ac, DX, xtot, 0.0_wp, res2, xtot/ac*ar)

  end subroutine krlm_old



  subroutine krrm(br, bc, xtot, neffpara, B, DX, res2)
    ! Does the multiplication kron(I,B)*DX
    !  We know that
    !    B -- br x bc
    !    DX -- xtot x neffpara
    !  We can deduce that
    !    X -- xr x xc
    !     xr = bc
    !     xc = xtot / xr
    !    I -- xc x xc
    !
    !  Because we are starting from vec(BdX)

    integer, intent(in) :: br, bc, xtot, neffpara

    real(wp), intent(in) :: B(br, bc), DX(xtot, neffpara)

    real(wp) :: res(br, neffpara*xtot/bc)

    real(wp), intent(out) :: res2(xtot/bc*br, neffpara)

    call dgemm('N', 'N', br, neffpara*xtot/bc, bc, 1.0_wp, B, br, DX, bc, 0.0_wp, res, br)

    res2 = reshape(res, (/xtot/bc*br, neffpara/))

  end subroutine krrm

  subroutine krrm_old(br, bc, xtot, neffpara, B, DX, res2)
    ! Does the multiplication kron(I,B)*DX
    !  We know that
    !    B -- br x bc
    !    DX -- xtot x neffpara
    !  We can deduce that
    !    X -- xr x xc
    !     xr = bc
    !     xc = xtot / xr
    !    I -- xc x xc
    !
    !  Because we are starting from vec(BdX)

    integer, intent(in) :: br, bc, xtot, neffpara

    real(wp), intent(in) :: B(br, bc), DX(xtot, neffpara)

    real(wp) :: res(br*bc, xtot), eye(xtot/bc,xtot/bc)

    real(wp), intent(out) :: res2(xtot/bc*br, neffpara)

    integer :: i

    eye = 0.0_wp
    do i = 1,(xtot/bc)
       eye(i,i) = 1.0_wp
    end do

    call Kronecker(xtot/bc, xtot/bc, br, bc, br*bc, xtot, 0.0_wp, 1.0_wp, eye, B, res)
    call dgemm('n','n',xtot/bc*br, neffpara, xtot, 1.0_wp, res, br*bc, DX, xtot, 0.0_wp, res2, xtot/bc*br)


  end subroutine krrm_old

  function commutation_matrix(m, n) result(C)
    
    integer, intent(in) :: m, n
    real(wp) :: C(m*n,m*n)

    integer :: i, j
    C = 0.0_wp
    do i = 1, m
       do j = 1, n
          C((i-1)*n + j, (j-1)*m + i)= 1.0_wp
       end do
    end do

  end function commutation_matrix


  subroutine utri2dens(X, nx)

    integer, intent(in) :: nx
    real(wp), intent(inout) :: X(nx,nx)

    integer :: i

    X = X + transpose(X)
    do i = 1,nx
       X(i,i) = 0.5_wp*X(i,i)
    end do

  end subroutine utri2dens


end module filter
